from IPython.display import display_html
from math import sqrt
from numpy import triu, ones, nan, inf
from pandas import DataFrame, Series, notna
from scipy.stats import kruskal
from statsmodels.formula.api import ols
from statsmodels.stats.outliers_influence import variance_inflation_factor


class FeatureSelector():
    """ A pipeline of measures to perform EDA and feature pre-selection"""
    
    def __init__(self, measures: list=[], filters: list=[]):
        
        self.measures = [dtype_measure, nans_measure, mode_measure] + measures[:]
        self.filters = [thresh_filter] + filters[:]
        
        self.associations = []
        self.filtered_associations = []
    
    def measure(self, x: Series, y: Series, **params):
        """ Measures association between x and y
        
        Parameters
        ----------
        thresh_nan, float: default 1.
          Maximum percentage of NaNs in a feature
        thresh_mode, float: default 1.
          Maximum percentage of the mode of a feature
        thresh_outlier, float: default 1.
          Maximum percentage of Outliers in a feature
        """
        
        passed = True
        association = {}
        for measure in self.measures:
            passed, association = measure(passed, association, x, y, **params)
            
        return association
    
    def measure_apply(self, X: DataFrame, y: Series, **params):
        """ Measures association between columns of X and y
        
        Parameters
        ----------
        thresh_nan, float: default 1.
          Maximum percentage of NaNs in a feature
        thresh_mode, float: default 1.
          Maximum percentage of the mode of a feature
        thresh_outlier, float: default 1.
          Maximum percentage of Outliers in a feature
        """
        
        # applying association measure to each column
        self.associations = X.apply(self.measure, y=y, **params, result_type='expand', axis=0).T
        
        return self.associations
    
    def filter_apply(self, X: DataFrame, sort_measure: str, **params):
        """ Filters out to correlated features that are less relevant
        
        Parameters
        ----------
        thresh_corr, float: default 1.
          Maximum association between features
        thresh_vif, float: default inf
          Maximum VIF between features
        ascending, bool: default False
          Ascending of Descending sort by sort_measure
        """
        
        # ordering features by sort_by
        self.filtered_associations = self.associations.sort_values(sort_measure, ascending=params.get('ascending', False))

        # applying successive filters
        for filtering in self.filters:

            # ordered filtering
            self.filtered_associations = filtering(X, self.filtered_associations, **params)
        
        return self.filtered_associations
    
    def display_stats(self, X: DataFrame, y: Series, sort_measure: str=None):
        """ Computes statistics for EDA with default params"""
        
        # initial computation of all association measures
        self.measure_apply(X, y)
        
        # filtering
        if not sort_measure:  # by default, last measure passed
            sort_measure = self.measures[-1].__name__
        association = self.filter_apply(X, sort_measure)
            
        # appllying style 
        subset = [c for c in association if 'pct_' in c or '_measure' in c]
        association = association.style.background_gradient(cmap='coolwarm', subset=subset)
        display_html(association.set_table_attributes("style='display:inline'")._repr_html_(), raw=True)
        
    
    def select_features(self, X: DataFrame, y: Series, n_best: int, sort_measures: list, **params):
        """ Selects the n_best features
        
        Parameters
        ----------
        thresh_nan, float: default 1.
          Maximum percentage of NaNs in a feature
        thresh_mode, float: default 1.
          Maximum percentage of the mode of a feature
        thresh_outlier, float: default 1.
          Maximum percentage of Outliers in a feature
        thresh_corr, float: default 1.
          Maximum association between features
        thresh_vif, float: default inf
          Maximum VIF between features
        ascending, bool: default False
          Ascending of Descending sort by sort_measure
        """
        
        # initial computation of all association measures
#         if len(self.associations) == 0:
        self.measure_apply(X, y, **params)
        
        # iterating over each sort_measures
        ranks = []
        for sort_measure in sort_measures:
            
            # filtering by sort_measure
            rank = list(self.filter_apply(X, sort_measure, **params).index)
            ranks += [rank]
        
        # retrieving the n_best features per ranking
        best_features = [feature for rank in ranks for feature in rank[:n_best]]
        best_features = list(set(best_features))  # deduplicating
        
        return best_features

    
# MEASURES
def nans_measure(active: bool, association: dict, x: Series, y: Series=None, **params):
    """ Measure of the percentage of NaNs"""
    
    passed = True  # whether or not tests where passed
    
    nans = x.isnull()  # ckecking for nans
    pct_nan = nans.mean()   # Computing percentage of nans
    
    # updating association
    association.update({'pct_nan': pct_nan})
    
    # Excluding feature that have to many NaNs
    passed = pct_nan < params.get('thresh_nan', 1.)
    
    return passed, association

def dtype_measure(active: bool, association: dict, x: Series, y: Series=None, **params):
    """ Gets dtype"""
    
    # updating association
    association.update({'dtype': x.dtype})
        
    return active, association

def mode_measure(active: bool, association: dict, x: Series, y: Series=None, **params):
    """ Measure of the percentage of the Mode"""
    
    passed = True  # whether or not tests where passed
    
    mode = x.mode(dropna=True).values[0]  # computing mode
    pct_mode = (x == mode).mean()  # Computing percentage of the mode
    
    # updating association
    association.update({'pct_mode': pct_mode, 'mode': mode})
    
    # Excluding feature with too frequent modes
    passed = pct_mode < params.get('thresh_mode', 1.)
    
    return passed, association

def kruskal_measure(active: bool, association: dict, x: Series, y: Series, **params):
    """ Kruskal-Wallis statistic between x (quantitative) and y (binary)"""
    
    # check that previous steps where passed
    if active:
    
        nans = x.isnull()  # ckecking for nans
        
        # computation of Kruskal-Wallis statistic
        kw = kruskal(x[(~nans) & (y == 0)], x[(~nans) & (y == 1)])
        
        # updating association
        if kw:
            association.update({'kruskal_measure': kw[0]})
        
    return active, association

def R_measure(active: bool, association: dict, x: Series, y: Series, **params):
    """ R of the linear regression of x (quantitative) by y (binary)"""
    
    # check that previous steps where passed
    if active:
    
        nans = x.isnull()  # ckecking for nans

        # grouping feature and target
        ols_df = DataFrame({'feature': x[~nans], 'target': y[~nans]})
        
        # fitting regression of feature by target
        regression = ols('feature~C(target)', ols_df).fit()
        
        # updating association
        if regression.rsquared:
            association.update({'R_measure': sqrt(regression.rsquared)})
        
    return active, association


def zscore_measure(active: bool, association: dict, x: Series, y: Series=None, **params):
    """ Computes outliers based on the z-score"""
    
    # check that previous steps where passed for computational optimization
    if active:
        
        mean = x.mean()  # mean of the feature
        std = x.std()  # standard deviation of the feature
        zscore = (x-mean) / std  # zscore per observation
        
        # checking for outliers
        outliers = abs(zscore) > 3
        pct_zscore = outliers.mean()
        
        # updating association
        association.update({
            'pct_zscore': pct_zscore,
            'min': x.min(),
            'max': x.max(),
            'mean': mean,
            'std': std
        })

        # Excluding feature with too frequent modes
        passed = pct_zscore < params.get('thresh_outlier', 1.)
        
    return active, association

def iqr_measure(active: bool, association: dict, x: Series, y: Series=None, **params):
    """ Computes outliers based on the inter-quartile range"""
    
    # check that previous steps where passed for computational optimization
    if active:
        
        q3 = x.quantile(0.75)  # 3rd quartile
        q1 = x.quantile(0.25)  # 1st quartile
        iqr = q3 - q1  # inter quartile range
        iqr_bounds = q1 - 1.5 * iqr, q3 + 1.5 * iqr  # bounds of the iqr range
        
        # checking for outliers
        outliers = ~x.between(*iqr_bounds)
        pct_iqr = outliers.mean()
        
        # updating association
        association.update({
            'pct_iqr': pct_iqr,
            'q1': q1,
            'median': x.median(),
            'q3': q3
        })

        # Excluding feature with too frequent modes
        passed = pct_iqr < params.get('thresh_outlier', 1.)
        
    return active, association

    
# FILTERS        
def thresh_filter(X: DataFrame, ranks: DataFrame, **params):
    """ Filters out missing association measure (did not pass a threshold)"""
    
    # drops rows with nans
    associations = ranks.dropna(axis=0)
    
    return associations

def quantitative_filter(X: DataFrame, ranks: DataFrame, corr_measure: str, **params):
    """ Computes max association between X and X (quantitative) excluding features 
    that are correlated to a feature more associated with the target 
    (defined by the prefered_order)."""
    
    # accessing the prefered order
    prefered_order = ranks.index

    # computing correlation between features
    X_corr = X[prefered_order].corr(corr_measure).abs()
    X_corr = X_corr.where(triu(ones(X_corr.shape), k=1).astype(bool))
    
    # initiating list of maximum association per feature
    associations = []
    
    # iterating over each feature by target association order
    for feature in prefered_order:
            
        # correlation with features more associated to the target
        corr_with_better_features = X_corr.loc[:feature, feature]
        
        # maximum correlation with a better feature
        corr_with, worst_corr = corr_with_better_features.agg(['idxmax', 'max'])
        
        # dropping the feature if it was too correlated to a better feature
        if worst_corr > params.get('thresh_corr', 1.):
            X_corr = X_corr.drop(feature, axis=0).drop(feature, axis=1)
            
        # kept feature: updating associations with this feature
        else:
            associations += [{
                'feature': feature, 
                f'{corr_measure}_measure': worst_corr,
                f'{corr_measure}_with': corr_with
            }]
            
    # formatting ouput to DataFrame
    associations = DataFrame(associations).set_index('feature')
            
    # applying filter on association
    associations = ranks.join(associations, how='right')
    
    return associations 


def spearman_filter(X: DataFrame, ranks: DataFrame, **params):
    """ Computes max Spearman between X and X (quantitative) excluding features 
    that are correlated to a feature more associated with the target 
    (defined by the prefered_order)."""
            
    # applying quantitative filter with spearman correlation
    associations = quantitative_filter(X, ranks, 'spearman', **params)
    
    return associations

def pearson_filter(X: DataFrame, ranks: DataFrame, **params):
    """ Computes max Pearson between X and X (quantitative) excluding features 
    that are correlated to a feature more associated with the target 
    (defined by the prefered_order)."""
            
    # applying quantitative filter with spearman correlation
    associations = quantitative_filter(X, ranks, 'pearson', **params)
    
    return associations

def vif_filter(X: DataFrame, ranks: DataFrame, **params):
    """ Computes Variance Inflation Factor (multicolinearity)"""
    
    # accessing the prefered order
    prefered_order = ranks.index
    
    # initiating list association per feature
    associations = []
    
    # list of dropped features
    dropped = []
    
    # iterating over each column
    for i, feature in enumerate(prefered_order):
        
        # identifying remaining more associated features
        better_features = [f for f in prefered_order[:i+1] if f not in dropped]
        
        X_vif = X[better_features]  # keeping only better features
        X_vif = X_vif.dropna(axis=0)  # dropping NaNs for OLS
        
        # computation of VIF
        vif = nan
        if len(better_features) > 1 and len(X_vif) > 0:
            vif = variance_inflation_factor(X_vif.values, len(better_features)-1)
        
        # dropping the feature if it was too correlated to a better feature
        if vif > params.get('thresh_vif', inf) and notna(vif):
            dropped += [feature]
            
        # kept feature: updating associations with this feature
        else:
            associations += [{'feature': feature, 'vif_measure': vif}]
    
    # formatting ouput to DataFrame
    associations = DataFrame(associations).set_index('feature')
            
    # applying filter on association
    associations = ranks.join(associations, how='right')
    
    return associations



from IPython.display import display_html
from numpy import sort, nan, inf, float16, where, isin, argsort, array, select, append, quantile, linspace, argmin
from pandas import DataFrame, Series, isna, qcut, notna, unique
from sklearn.base import BaseEstimator, TransformerMixin
from warnings import warn


class GroupedList(list):
    
    def __init__(self, iterable=()):
        """ An ordered list that historizes its elements' merges."""
        
        # case 0: iterable est le dictionnaire contained
        if isinstance(iterable, dict):
            
            # récupération des valeurs de la liste (déjà ordonné)
            values = [key for key, _ in iterable.items()]

            # initialsiation de la liste
            super().__init__(values)

            # attribution des valeurs contenues
            self.contained = {k: v for k, v in iterable.items()}
        
        # case 1: s'il ne s'agit pas déjà d'une liste groupée -> création des groupes
        elif isinstance(iterable, GroupedList):

            # initialsiation de la liste
            super().__init__(iterable)

            # copie des groupes
            self.contained = {k: v for k, v in iterable.contained.items()}
        
        # case 2: il s'agit d'une GroupedList -> copie des groupes
        else:

            # initialsiation de la liste
            super().__init__(iterable)

            # création des groupes
            self.contained = {v: [v] for v in iterable}

    def group_list(self, to_discard: list, to_keep: str):
        """ Groups elements to_discard into values to_keep"""

        for discarded, kept in zip(to_discard, [to_keep] * len(to_discard)):
            self.group(discarded, kept)

    def group(self, discarded, kept):
        """ Groups the discarded value with the kept value"""

        if not is_equal(discarded, kept):

            assert discarded in self, f"{discarded} not in list"
            assert kept in self, f"{kept} not in list"

            contained_discarded = self.contained.get(discarded)
            contained_kept = self.contained.get(kept)

            self.contained.update({
                kept: contained_discarded + contained_kept,
                discarded: []
            })

            self.remove(discarded)
        
        return self
        
    def append(self, new_value):
        """ Appends a new_value to the GroupedList"""
        
        self += [new_value]
        
        self.contained.update({new_value: [new_value]})
        
        return self

    def sort(self):
        """ Sorts the values of the list and dict, if any, NaNs are the last. """

        # str values
        keys_str = [key for key in self if isinstance(key, str)]

        # non-str values
        keys_float = [key for key in self if not isinstance(key, str)]

        # sorting and merging keys
        keys = list(sort(keys_str)) + list(sort(keys_float)) 

        # recreating an ordered GroupedList
        self = GroupedList({key: self.get(key) for key in keys})

        return self

    def sort_by(self, ordering):
        """ Sorts the values of the list and dict, if any, NaNs are the last. """

        # recreating an ordered GroupedList
        self = GroupedList({k: self.get(k) for k in ordering})

        return self

    
    def remove(self, value):
        
        super().remove(value)
        self.contained.pop(value)
    
    def pop(self, idx):
        
        value = self[idx]
        self.remove(value)
    
    def get(self, key):
        """ returns list of values contained in key"""

        # default to fing an element
        found = self.contained.get(key)

        # copying with dictionnaries (not working with numpy.nan)
        if isna(key):
            found = [value for dict_key, value in self.contained.items() if is_equal(dict_key, key)][0]

        return found

    def get_group(self, value):
        """ returns the group containing the specified value """
        
        found = [key for key, values in self.contained.items() if any(is_equal(value, elt) for elt in values)]

        if any(found):
            return found[0]

    def values(self):
        """ returns all values contained in each group """

        known = [value for values in self.contained.values() for value in values]

        return known

    def contains(self, value):
        """ checks if a value if contained in any group """

        known_values = self.values()

        return any(is_equal(value, known) for known in known_values)


class Discretizer(BaseEstimator, TransformerMixin):
    """ Automatic discretizing of continuous, categorical and categorical ordinal features.

    Modalities/values of features are grouped according to there respective orders:
     - [Qualitative features] order based on modality target rate.
     - [Qualitative ordinal features] user-specified order.
     - [Quantitative features] real order of the values.

    Parameters
    ----------
    quanti_features: list, default []
        Contains quantitative (continuous) features to be discretized.

    quali_features: list, default []
        Contains qualitative (categorical and categorical ordinal) features to be discretized.

    min_freq: int, default None
        [Qualitative features] Minimal frequency of a modality.
         - NaNs are considered a specific modality but will not be grouped.
         - [Qualitative features] Less frequent modalities are grouped in the `__OTHER__` modality.
         - [Qualitative ordinal features] Less frequent modalities are grouped to the closest modality 
        (smallest frequency or closest target rate), between the superior and inferior values (specified
        in the `values_orders` dictionnary).
        Recommandation: `min_freq` should be set from 0.01 (preciser) to 0.05 (faster, increased stability).

    q: int, default None
        [Quantitative features] Number of quantiles to initialy cut the feature.
         - NaNs are considered a specific value but will not be grouped.
         - Values more frequent than `1/q` will be set as their own group and remaining frequency will be
        cut into proportionaly less quantiles (`q:=max(round(non_frequent * q), 1)`). 
        Exemple: if q=10 and the value numpy.nan represent 50 % of the observed values, non-NaNs will be 
        cut in q=5 quantiles.
        Recommandation: `q` should be set from 10 (faster) to 20 (preciser).

    values_orders: dict, default {}
        [Qualitative ordinal features] dict of features values and list of orders of their values.
         - [Qualitative ordinal features] Less frequent modalities are grouped to the closest modality 
        (smallest frequency or closest target rate), between the superior and inferior values (described
        by the `values_orders`).
        Exemple: for an `age` feature, `values_orders` could be `{'age': ['0-18', '18-30', '30-50', '50+']}`.
    """
    
    def __init__(self, quanti_features: list=[], quali_features: list=[], q: int=None, min_freq: float=None, values_orders: dict={}, copy: bool=False, verbose: bool=False):
    
        
        self.features = quanti_features[:] + quali_features[:]
        self.quanti_features = quanti_features[:]
        self.quali_features = quali_features[:]
        self.values_orders = {k: GroupedList(v) for k, v in values_orders.items()}
        self.min_freq = min_freq
        self.q = q
        self.pipe = []
        self.copy = copy
        self.verbose = verbose

    def fit(self, X, y):
        
        # copie des features
        Xc = X.copy()

        # [Qualitative ordinal features] Grouping rare values into closest common one
        if any(self.values_orders.keys()):
            if self.verbose: print("\n---\n[Discretizer] Fitting Qualitative Ordinal Features")

            discretizer = ClosestDiscretizer(self.values_orders, min_freq=self.min_freq, verbose=self.verbose)
            discretizer.fit(Xc, y)

            # storing results
            self.values_orders.update(discretizer.values_orders)  # adding orders of grouped features
            self.pipe += [('QualiClosestDiscretizer', discretizer)]  # adding discretizer to pipe

        # [Qualitative non-ordinal features] Grouping rare values into default_value '__OTHER__'
        non_ordinal = [f for f in self.quali_features if f not in self.values_orders.keys()]
        if any(non_ordinal):
            if self.verbose: print("\n---\n[Discretizer] Fitting Qualitative Non-Ordinal Features")

            discretizer = DefaultDiscretizer(non_ordinal, min_freq=self.min_freq, verbose=self.verbose)
            discretizer.fit(Xc, y)

            # storing results
            self.values_orders.update(discretizer.values_orders)  # adding orders of grouped features
            self.pipe += [('DefaultDiscretizer', discretizer)]  # adding discretizer to pipe

        # [Quantitative features] Grouping values into quantiles
        if any(self.quanti_features):
            if self.verbose: print("\n---\n[Discretizer] Fitting Quantitative Features")

            discretizer = QuantileDiscretizer(self.quanti_features, q=self.q, verbose=self.verbose)
            Xc = discretizer.fit_transform(Xc)

            # storing results
            self.values_orders.update(discretizer.values_orders)  # adding orders of grouped features
            self.pipe += [('QuantileDiscretizer', discretizer)]  # adding discretizer to pipe

            # [Quantitative features] Grouping rare quantiles into closest common one 
            #  -> can exist because of overrepresented values (values more frequent than 1/q)
            frequencies = Xc[self.quanti_features].apply(lambda u: u.value_counts(dropna=False, normalize=True).drop(nan, errors='ignore').reindex(self.values_orders.get(u.name)).fillna(0).values.min(), axis=0)  # computing min frequency per quantitative feature
            q_min_freq = 1 / self.q / 2
            has_rare = list(frequencies[frequencies <= q_min_freq].index)
            if any(has_rare):
                if self.verbose: print("\n---\n[Discretizer] Fitting Quantitative Features (that have rare values)")

                rare_values_orders = {feature: order for feature, order in self.values_orders.items() if feature in has_rare}
                discretizer = ClosestDiscretizer(rare_values_orders, min_freq=q_min_freq, verbose=self.verbose)
                discretizer.fit(Xc, y)

                # storing results
                self.values_orders.update(discretizer.values_orders)  # adding orders of grouped features
                self.pipe += [('QuantiClosestDiscretizer', discretizer)]  # adding discretizer to pipe

        return self

    def transform(self, X, y=None):
        if self.verbose: print("\n---\n[Discretizer] Discretizing Features")
        
        # copie des features
        Xc = X
        if self.copy:
            Xc = X.copy()

        # iterating over each transformer
        for _, step in self.pipe:
            Xc = step.transform(Xc)

        return Xc


class GroupedListDiscretizer(BaseEstimator, TransformerMixin):
    
    def __init__(self, values_orders: dict, *, copy: bool=False, output: type= float, verbose: bool=False):
        
        self.features = list(values_orders.keys())
        self.values_orders = {feature: GroupedList(order) for feature, order in values_orders.items()}
        self.copy = copy
        self.output = output
        self.verbose = verbose
        
    def fit(self, X, y=None):

        return self
    
    def transform(self, X, y=None):

        # copying dataframes
        Xc = X
        if self.copy:
            Xc = X.copy()

        # iterating over each feature
        for n, feature in enumerate(self.features):
            if self.verbose: print(f" - [GroupedListDiscretizer] Discretizing {feature} ({n+1}/{len(self.features)})")
            
            order = self.values_orders.get(feature)  # récupération des groupes
            to_discard = [order.get(group) for group in order]  # identification des valeur à regrouper
            to_input = [Xc[feature].isin(discarded) for discarded in to_discard]  # identification des observations à regrouper
            to_keep = [n if self.output == float else group for n, group in enumerate(order)]  # récupération du groupe dans lequel regrouper
            Xc[feature] = select(to_input, to_keep, default=Xc[feature])  # grouping modalities

        # converting to float
        if self.output == float:
            Xc[self.features] = Xc[self.features].astype(float16)

        return Xc

class ChainedDiscretizer(GroupedListDiscretizer):
    
    def __init__(self, features: list, min_freq: float, chained_orders: list, *, copy: bool=False, verbose: bool=False):       
        
        self.min_freq = min_freq
        self.features = features[:]
        self.chained_orders = [GroupedList(order) for order in chained_orders]
        self.pipe = []
        self.copy = copy
        self.verbose = verbose
        
        # initiating features' values orders to all possible values
        self.known_values = list(set([value for group in self.chained_orders for value in group.values()]))
        self.values_orders = {f: GroupedList(self.known_values[:]) for f in self.features}

    def fit(self, X, y=None):
        
        # copying dataframe
        Xc = X[self.features].copy()
        
        # iterating over each feature
        for n, feature in enumerate(self.features):
            if self.verbose: print(f" - [ChainedDiscretizer] Fitting {feature} ({n+1}/{len(self.features)})")

            # computing frequencies of each modality
            frequencies = Xc[feature].value_counts(dropna=False, normalize=True).drop(nan, errors='ignore')  # dropping nans to keep them anyways
            values, frequencies = frequencies.index, frequencies.values

            # checking for unknown values (values to present in an order of self.chained_orders)
            missing = [value for value in values if notna(value) and (value not in self.known_values)]
            assert not any(missing), f"Order needs to be provided for values: {missing}"

            # iterating over each specified orders
            for order in self.chained_orders:
                
                # identifying modalities which rarest values
                to_keep = values[frequencies >= self.min_freq]

                # grouping rare modalities
                to_discard = [[value for value in order.get(group) if (not value in to_keep)] for group in order]  # identifying rare values
                to_input = [Xc[feature].isin(discarded) for discarded in to_discard]  # identifying observation to input
                Xc[feature] = select(to_input, to_input, default=Xc[feature])  # regroupement des naf peu fréquents
                
                # historizing in the feature's order
                for discarded, kept in zip(to_discard, order):
                    self.values_orders.get(feature).group_list(discarded, kept)  # historisation dans l'order
                    
                # updating frequencies of each modality for the next ordering
                frequencies = Xc[feature].value_counts(dropna=False, normalize=True).drop(nan, errors='ignore')  # dropping nans to keep them anyways
                values, frequencies = frequencies.index, frequencies.values
        
        super().__init__(self.values_orders, copy=self.copy, output=str)
        super().fit(X, y)
            
        return self

class QuantileDiscretizer(BaseEstimator, TransformerMixin):
    
    def __init__(self, features: list, q: int, copy: bool=False, verbose: bool=False):
        """ Discretizes quantitative features into groups of q quantiles"""
        
        self.features = features[:]
        self.q = q
        self.values_orders = dict()
        self.quantiles = None
        self.copy = copy
        self.verbose = verbose

    def fit(self, X, y=None):
        
        # computing quantiles for the feature
        self.quantiles = X[self.features].apply(find_quantiles, q=self.q, axis=0)

        # case when only one feature is discretized
        if len(self.features) == 1:
        	self.quantiles = {self.features[0]: list(self.quantiles.get(self.features[0]).values)}

        # building string of values to be displayed
        values = []
        for n, feature in enumerate(self.features):
            if self.verbose: print(f" - [QuantileDiscretizer] Fitting {feature} ({n+1}/{len(self.features)})")

            # quantiles as strings
            feature_quantiles = self.quantiles.get(feature)
            str_values = ['<= ' + q for q in format_list(feature_quantiles)]

            # case when there is only one value
            if len(feature_quantiles) == 0:
                str_values = ['non-nan']

            # last quantile is between the last value and inf
            else:
                str_values = str_values + [str_values[-1].replace('<= ', '>  ')]
            values += [str_values]
        
        # adding inf for the last quantiles
        self.quantiles = {f: q + [inf] for f, q in self.quantiles.items()}

        # creating the values orders based on quantiles
        self.values_orders.update({feature: GroupedList(str_values) for feature, str_values in zip(self.features, values)})

        return self
    
    def transform(self, X, y=None):
        
        # copying dataset if requested
        Xc = X
        if self.copy:
            Xc = X.copy()
        
        # iterating over each feature
        for n, feature in enumerate(self.features):
            if self.verbose: print(f" - [QuantileDiscretizer] Discretizing {feature} ({n+1}/{len(self.features)})")

            nans = isna(Xc[feature])  # keeping track of nans

            # grouping values inside quantiles
            to_input = [Xc[feature] <= q for q in self.quantiles.get(feature)]  # values that will be imputed
            values = [[v] * len(X) for v in self.values_orders.get(feature)]  # new values to imput
            Xc[feature] = select(to_input, values, default=Xc[feature])  # grouping modalities

            # adding back nans
            if any(nans):
                Xc.loc[nans, feature] = nan
        
        return Xc

class ClosestDiscretizer(BaseEstimator, TransformerMixin):
    
    def __init__(self, values_orders: dict, min_freq: float, *, default: str='worst', copy: bool=False, verbose: bool=False):
        """ Discretizes ordered qualitative features into groups more frequent than min_freq"""
        
        self.features = list(values_orders.keys())
        self.min_freq = min_freq
        self.values_orders = {k: GroupedList(v) for k, v in values_orders.items()}
        self.default = default[:]
        self.default_values = {}
        self.copy = copy
        self.verbose = verbose

    def fit(self, X, y):
        
        # grouping rare modalities for each feature
        common_modalities = X[self.features].apply(
            lambda u: find_common_modalities(u, y, self.min_freq, self.values_orders.get(u.name)), 
            axis=0, result_type='expand'
        ).T.to_dict()

        # updating the order per feature
        self.values_orders = {f: common_modalities.get('order').get(f) for f in self.features}

        # defining the default value based on the strategy
        self.default_values = {f: common_modalities.get(self.default).get(f) for f in self.features}

        return self
    
    def transform(self, X, y=None):
        
        # copying dataset if requested
        Xc = X
        if self.copy:
            Xc = X.copy()

        # iterating over each feature
        for n, feature in enumerate(self.features):
            if self.verbose: print(f" - [ClosestDiscretizer] Discretizing {feature} ({n+1}/{len(self.features)})")

            # accessing feature's modalities' order
            order = self.values_orders.get(feature)

            # imputation des valeurs inconnues le cas échéant
            unknowns = [value for value in Xc[feature].unique() if not any(is_equal(value, known) for known in order.values())]
            unknowns = [value for value in unknowns if notna(value)]  # suppression des NaNs
            if any(unknowns):
                to_input = [Xc[feature] == unknown for unknown in unknowns]  # identification des valeurs à regrouper
                Xc[feature] = select(to_input, [self.default_values.get(feature)], default=Xc[feature])  # regroupement des valeurs
                warn(f"Unknown modalities provided for {feature}: {unknowns}")

            # grouping values inside groups of modalities
            to_discard = [order.get(group) for group in order]  # identification des valeur à regrouper
            to_input = [Xc[feature].isin(discarded) for discarded in to_discard]  # identification des valeurs à regrouper
            Xc[feature] = select(to_input, order, default=Xc[feature])  # regroupement des valeurs

        return Xc


class DefaultDiscretizer(BaseEstimator, TransformerMixin):
    
    def __init__(self, features: list, min_freq: float, *, default_value: str='__OTHER__', copy: bool=False, verbose: bool=False):
        """ Groups a qualitative features' values less frequent than min_freq into a default_value"""
        
        self.features = features[:]
        self.min_freq = min_freq
        self.values_orders = {}
        self.default_value = default_value[:]
        self.copy = copy
        self.verbose = verbose

    def fit(self, X, y):

        # computing frequencies of each modality
        frequencies = X[self.features].apply(lambda u: u.value_counts(dropna=False, normalize=True).to_dict(), axis=0).to_dict()

        # computing target rate per modality for ordering
        target_rates = X[self.features].apply(lambda u: y.groupby(u, dropna=True).mean().sort_values().to_dict(), axis=0).to_dict()

        # attributing orders to each feature
        self.values_orders = {feature: GroupedList([value for value, _ in target_rates.get(feature).items()]) for feature in self.features}

        # identifying modalities which are the most common
        self.to_keep = {feature: [value for value, frequency in frequencies.get(feature).items() if (frequency >= self.min_freq) & notna(value)] + [nan] for feature in self.features}

        # grouping rare modalities 
        for n, feature in enumerate(self.features):
            if self.verbose: print(f" - [DefaultDiscretizer] Fitting {feature} ({n+1}/{len(self.features)})")
            
            # identifying values to discard (rare modalities)
            to_discard = [value for value in self.values_orders.get(feature) if value not in self.to_keep.get(feature)]
            
            # discarding rare modalities
            if any(to_discard):
                
                # adding default_value to possible values
                order = self.values_orders.get(feature)
                order.append(self.default_value)
                
                # grouping rare modalities into default_value
                order.group_list(to_discard, self.default_value)
                
                # computing target rate for default value and reordering values according to feature's target rate
                default_target_rate = y.loc[X[feature].isin(order.get(self.default_value))].mean()  # computing target rate for default value
                order_target_rate = [target_rates.get(feature).get(value) for value in order if value != self.default_value]
                default_position = next(n for n, trate in enumerate(order_target_rate + [inf]) if trate > default_target_rate)
                new_order = order[:-1][:default_position] + [self.default_value] + order[:-1][default_position:]  # getting rid of default value already in order
                order = order.sort_by(new_order)
                self.values_orders.update({feature: order})

        return self

    def transform(self, X, y=None):

        # copying dataset if requested
        Xc = X
        if self.copy:
            Xc = X.copy()

        # grouping values inside groups of modalities
        for n, feature in enumerate(self.features):
            if self.verbose: print(f" - [DefaultDiscretizer] Discretizing {feature} ({n+1}/{len(self.features)})")

            # grouping modalities
            Xc[feature] = select([~Xc[feature].isin(self.to_keep.get(feature))], [self.default_value], default=Xc[feature])

        return Xc

def find_quantiles(df_feature: Series, q: int, len_df: int=None, quantiles: list=None):
    """ Découpage en quantile de la feature.
    
    Fonction récursive : on prend l'échantillon et on cherche des valeurs sur-représentées
    Si la valeur existe on la met dans une classe et on cherche à gauche et à droite de celle-ci, d'autres variables sur-représentées
    Une fois il n'y a plus de valeurs qui soient sur-représentées,
    on fait un découpage classique avec qcut en multipliant le nombre de classes souhaité par le pourcentage de valeurs restantes sur le total
    
    """

    # initialisation de la taille total du dataframe
    if len_df is None:
        len_df = len(df_feature)
    
    # initialisation de la liste des quantiles
    if quantiles is None:
        quantiles = []

    # calcul du nombre d'occurences de chaque valeur
    frequencies = df_feature.value_counts(dropna=False, normalize=False).drop(nan, errors='ignore') / len_df # dropping nans to keep them anyways
    values, frequencies = array(frequencies.index), array(frequencies.values)
    
    # cas 1 : il n'y a pas d'observation dans le dataframe
    if len(df_feature) == 0:
        
        return quantiles
    
    # cas 2 : il y a des valeurs manquantes NaN
    elif any(isna(df_feature)):
        
        return find_quantiles(df_feature[notna(df_feature)], q, len_df=len_df, quantiles=quantiles)
        
    # cas 2 : il n'y a que des valeurs dans le dataframe (pas de NaN)
    else:
        
        # cas 1 : il existe une valeur sur-représentée
        if any(frequencies > 1 / q):

            # identification de la valeur sur-représentée
            frequent_value = values[frequencies.argmax()]
            
            # ajout de la valeur fréquente à la liste des quantiles
            quantiles += [frequent_value]

            # calcul des quantiles pour les parties inférieures et supérieures
            quantiles_inf = find_quantiles(df_feature[df_feature < frequent_value], q, len_df=len_df)
            quantiles_sup = find_quantiles(df_feature[df_feature > frequent_value], q, len_df=len_df)
            
            return quantiles_inf + quantiles + quantiles_sup

        # cas 2 : il n'existe pas de valeur sur-représentée
        else:
            
            # nouveau nombre de quantile en prenant en compte les classes déjà constituées
            new_q = max(round(len(df_feature) / len_df * q), 1)
            
            # calcul des quantiles sur le dataframe
            if new_q > 1:
                quantiles += list(quantile(df_feature.values, linspace(0, 1, new_q + 1)[1:-1], interpolation='lower'))

            # case when there are no enough observations to compute quantiles
            else:
                quantiles += [max(df_feature.values)]

            return quantiles

def is_equal(a, b):
    """ checks if a and b are equal (NaN insensitive)"""
    
    # default equality
    equal = a == b
    
    # Case where a and b are NaNs
    if isna(a) and isna(b):
        equal = True
    
    return equal

def find_common_modalities(df_feature: Series, y: Series, min_freq: float, order: GroupedList, len_df: int=None):
    """ Découpage en modalités de fréquence minimal (Cas des variables ordonnées).
    
    Fonction récursive : on prend l'échantillon et on cherche des valeurs sur-représentées
    Si la valeur existe on la met dans une classe et on cherche à gauche et à droite de celle-ci, d'autres variables sur-représentées
    Une fois il n'y a plus de valeurs qui soient sur-représentées,
    on fait un découpage classique avec qcut en multipliant le nombre de classes souhaité par le pourcentage de valeurs restantes sur le total
    
    """

    # initialisation de la taille totale du dataframe
    if len_df is None:
        len_df = len(df_feature)
    
    # conversion en GroupedList si ce n'est pas le cas 
    order = GroupedList(order)
    
    # cas 1 : il y a des valeurs manquantes NaN
    if any(isna(df_feature)):
        
        return find_common_modalities(df_feature[notna(df_feature)], y[notna(df_feature)], min_freq, order, len_df)


    # cas 2 : il n'y a que des valeurs dans le dataframe (pas de NaN)
    else:
        
        # computing frequencies and target rate of each modality
        init_frequencies = df_feature.value_counts(dropna=False, normalize=True).drop(nan, errors='ignore')  # dropping nans to keep them anyways
        init_values, init_frequencies = init_frequencies.index, init_frequencies.values
        
        # ordering
        frequencies = [init_frequencies[init_values == value][0] if any(init_values == value) else 0 for value in order]  # sort selon l'ordre des modalités
        values = [init_values[init_values == value][0] if any(init_values == value) else value for value in order]  # sort selon l'ordre des modalités
        target_rate = y.groupby(df_feature).sum().reindex(order) / frequencies / len_df # target rate per modality
        underrepresented = [value for value, frequency in zip(values, frequencies) if frequency < min_freq]  # valeur peu fréquentes

        # cas 1 : il existe une valeur sous-représentée
        while any(underrepresented) & (len(frequencies) > 1):

            # identification de la valeur sous-représentée
            discarded_idx = argmin(frequencies)
            discarded = values[discarded_idx]

            # identification de la modalité la plus proche (volume et taux de défaut)
            kept = find_closest_modality(discarded, discarded_idx, list(zip(order, frequencies, target_rate)), min_freq)

            # removing the value from the initial ordering
            order.group(discarded, kept)
            
            # ordering
            frequencies = [init_frequencies[init_values == value][0] if any(init_values == value) else 0 for value in order]  # sort selon l'ordre des modalités
            values = [init_values[init_values == value][0] if any(init_values == value) else value for value in order]  # sort selon l'ordre des modalités
            target_rate = y.groupby(df_feature).sum().reindex(order) / frequencies / len_df # target rate per modality
            underrepresented = [value for value, frequency in zip(values, frequencies) if frequency < min_freq]  # valeur peu fréquentes

        worst, best = target_rate.idxmin(), target_rate.idxmax()

        # cas 2 : il n'existe pas de valeur sous-représentée
        return {'order': order, 'worst': worst, 'best': best}

def find_closest_modality(value, idx, freq_target, min_freq):
    """HELPER Finds the closesd modality in terms of frequency and target rate"""

    # cas 1: il s'agit de la modalité la plus petite
    if idx == 0:
        closest_modality = 1
        
    # cas 2: il s'agit de la modalité la plus grande
    elif idx == len(freq_target) - 1:
        closest_modality = len(freq_target) - 2
    
    # cas 3: il s'agit d'un cas intermédiaire
    else:
        # previous modality's volume and target rate
        previous_value, previous_volume, previous_target = freq_target[idx - 1]
        
        # current modality's volume and target rate
        _, volume, target = freq_target[idx]
        
        # next modality's volume and target rate
        next_value, next_volume, next_target = freq_target[idx + 1]
        
        # regroupement par volumétrie (préféré s'il n'y a pas beaucoup de volume)
        # case 1: la modalité suivante est inférieure à min_freq 
        if next_volume < min_freq <= previous_volume:
            closest_modality = idx + 1
            
        # case 2: la modalité précédante est inférieure à min_freq 
        elif previous_volume < min_freq <= next_volume:
            closest_modality = idx - 1
            
        # case 3: les deux modalités sont inférieures à min_freq 
        elif (previous_volume < min_freq) & (next_volume < min_freq):
            
            # cas 1: la modalité précédante est inférieure à la modalité suivante
            if previous_volume < next_volume:
                closest_modality = idx - 1
            
            # cas 2: la modalité suivante est inférieure à la modalité précédante
            elif next_volume < previous_volume:
                closest_modality = idx + 1
            
            # cas 3: les deux modalités ont la même fréquence
            else:
                
                # cas1: la modalité précédante est plus prorche en taux de cible
                if abs(previous_target - target) <= abs(next_target - target):
                    closest_modality = idx - 1
                    
                # cas2: la modalité précédante est plus prorche en taux de cible
                else:
                    closest_modality = idx + 1
                
        # case 4: les deux modalités sont supérieures à min_freq 
        else:
                
            # cas1: la modalité précédante est plus prorche en taux de cible
            if abs(previous_target - target) <= abs(next_target - target):
                closest_modality = idx - 1

            # cas2: la modalité précédante est plus prorche en taux de cible
            else:
                closest_modality = idx + 1
    
    # récupération de la valeur associée
    closest_value = freq_target[closest_modality][0]
    
    return closest_value


def format_list(a_list: list):
    """ Formats a list of floats to a list of unique rounded strings of floats"""

    # finding the closest power of thsousands for each element
    closest_powers = [next((k for k in range(-3, 4) if abs(elt) // 1000**(k) < 10)) for elt in a_list]

    # rounding elements to the closest power of thousands
    rounded_to_powers = [elt / 1000**(k) for elt, k in zip(a_list, closest_powers)]

    # computing optimal decimal per unique power of thousands
    optimal_decimals = {}
    for power in unique(closest_powers):  # iterating over each power of thousands found

        # filtering on the specific power of thousands
        sub_array = array([elt for elt, p in zip(rounded_to_powers, closest_powers) if power == p])

        # number of distinct values
        n_uniques = sub_array.shape[0]

        # computing the first rounding decimal that allows for distinction of each values when rounded
        # by default None (no rounding)
        optimal_decimal = next((k for k in range(1, 10) if len(unique(sub_array.round(k))) == n_uniques), None)

        # storing in the dict    
        optimal_decimals.update({
            power: optimal_decimal
        })

    # rounding according to each optimal_decimal
    rounded_list = []
    for elt, power in zip(rounded_to_powers, closest_powers):

        # rounding each element
        rounded = elt  # by default None (no rounding)
        optimal_decimal = optimal_decimals.get(power)
        if optimal_decimal:  # checking that the optimal decimal exists
            rounded = round(elt, optimal_decimal)

        # adding the rounded number to the list
        rounded_list += [rounded]

    # dict of suffixes per power of thousands
    suffixes = {
        -3: 'n', -2: 'mi', -1: 'm', 0: '', 1: 'K', 2: 'M', 3: 'B'
    }

    # adding the suffixes
    formatted_list = [str(elt) + suffixes.get(power) for elt, power in zip(rounded_list, closest_powers)]

    return formatted_list


import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name='AutoCarver',
    version='3.0.9',
    author='Mario DEFRANCE',
    author_email='defrancemario@gmail.com',
    description='Automatic Carving of Features with Optimal Association',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/mdefrance/AutoCarver',
    project_urls = {
        "Bug Tracker": "https://github.com/mdefrance/AutoCarver/issues"
    },
    license='MIT',
    packages=['AutoCarver']
)



from .Discretizers import GroupedList, GroupedListDiscretizer, is_equal
from IPython.display import display_html
from matplotlib.pyplot import subplots
from matplotlib.ticker import PercentFormatter
from numpy import sort, nan, inf, float32, where, isin, argsort, array, append, quantile, linspace, argmin, sqrt, random
from pandas import DataFrame, Series, isna, qcut, notna, unique, concat, crosstab
from scipy.stats import chi2_contingency
from seaborn import color_palette, despine
from tqdm.notebook import tqdm


class AutoCarver(GroupedListDiscretizer):
    """ Automatic carving of continuous, categorical and categorical ordinal features 
    that maximizes association with a binary target.

    Modalities/values of features are carved/regrouped according to a computed specific
    order defined based on their types:
     - [Qualitative features] grouped based on modality target rate.
     - [Qualitative ordinal features] grouped based on specified modality order.
     - [Quantitative features] grouped based on the order of their values.
    Uses Tschurpow's T to find the optimal carving (regrouping) of modalities/values
    of features.

    Parameters
    ----------
    values_orders: dict, default {}
        Dictionnary of features and list of their respective values' order.
        Exemple: for an `age` feature, `values_orders` could be `{'age': ['0-18', '18-30', '30-50', '50+']}`.

    sort_by: str, default 'tschuprowt'
        Association measure used to find the optimal group modality combination.
        Implemented: ['cramerv', 'tschuprowt']

    max_n_mod: int, default 5
        Maximum number of modalities for the carved features (excluding `numpy.nan`).
         - All possible combinations of less than `max_n_mod` groups of modalities will be tested. 
        Recommandation: `max_n_mod` should be set from 4 (faster) to 6 (preciser).

    keep_nans: bool, default False
        Whether or not to group `numpy.nan` to other modalities/values.

    Examples
    ----------

    from AutoCarver import AutoCarver
    from Discretizers import Discretizer
    from sklearn.pipeline import Pipeline

    # defining training and testing sets
    X_train, y_train = train_set, train_set[target]
    X_test, y_test = test_set, test_set[target]

    # specifying features to be carved
    selected_quanti = ['amount', 'distance', 'length', 'height']
    selected_quali = ['age', 'type', 'grade', 'city']

    # specifying orders of categorical ordinal features
    values_orders = {
        'age': ['0-18', '18-30', '30-50', '50+'],
        'grade': ['A', 'B', 'C', 'D', 'J', 'K', 'NN']
    }

    # pre-processing of features into categorical ordinal features
    discretizer = Discretizer(selected_quanti, selected_quali, min_freq=0.02, q=20, values_orders=values_orders)
    X_train = discretizer.fit_transform(X_train, y_train)
    X_test = discretizer.transform(X_test)

    # storing Discretizer
    pipe = [('Discretizer', discretizer)]

    # updating features' values orders (at this step every features are qualitative ordinal)
    values_orders = discretizer.values_orders

    # intiating AutoCarver
    auto_carver = AutoCarver(values_orders, sort_by='cramerv', max_n_mod=5, sample_size=0.01)

    # fitting on training sample
    # a test sample can be specified to evaluate carving robustness
    X_train = auto_carver.fit_transform(X_train, y_train, X_test, y_test)

    # applying transformation on test sample
    X_test = auto_carver.transform(X_test)

    # identifying non stable/robust features
    print(auto_carver.non_viable_features)

    # storing fitted GroupedListDiscretizer in a sklearn.pipeline.Pipeline
    pipe += [('AutoCarver', auto_carver)]
    pipe = Pipeline(pipe)

    # applying pipe to a validation set or in production
    X_val = pipe.transform(X_val)

    """
    
    def __init__(self, values_orders: dict={}, *, sort_by: str='tschuprowt',
                 copy: bool=False, max_n_mod: int=5, keep_nans: bool=False,
                 verbose: bool=True):
        
        self.features = list(values_orders.keys())
        self.values_orders = {k: GroupedList(v) for k, v in values_orders.items()}
        self.target = None  # le nom de la target est récupéré lors du fit
        self.non_viable_features = []  # liste des features pas stable entre TRAIN et TEST
        self.max_n_mod = max_n_mod  # nombre maximal de modalités
        self.keep_nans = keep_nans  # whether or not to group NaNs with other modalities
        self.sort_by = sort_by  # association measure used find the best groups
        self.copy = copy
        self.verbose = verbose

    def fit(self, X: DataFrame, y: Series, X_test: DataFrame=None, y_test: Series=None):
 
        # copying dataframes
        Xc = X
        Xtestc = X_test if X_test is not None else DataFrame(columns=X.columns)
        if self.copy:
            Xc = X.copy()
            Xtestc = X_test.copy() if X_test is not None else DataFrame(columns=X.columns)

        # récupération du nom de la target
        self.target = y.name

        # découpage automatique de chacune des variables
        for n, feature in tqdm(enumerate(self.features), total=len(self.features), disable=self.verbose):

            # printing the group statistics and determining default ordering
            if self.verbose: print(f"\n---\n[AutoCarver] Fitting {feature} ({n+1}/{len(self.features)})")

            # getting best combination
            best_groups = self.get_best_combination(feature, X, y, X_test, y_test)

            # feature can not be carved robustly
            if not bool(best_groups):
                self.non_viable_features += [feature]
                print(f"Feature {feature} is not viable")

        # discretizing features based on each feature's values_order
        super().__init__(self.values_orders, copy=self.copy, output=float)
        super().fit(X, y)

        return self

    def get_best_combination(self, feature: str, X: DataFrame, y: Series, X_test: DataFrame, y_test: Series):

        # getting all possible combinations for the feature without NaNS
        combinations = get_all_combinations(self.values_orders.get(feature), X[feature].nunique(dropna=True), self.max_n_mod)
        combinations = [GroupedList({group[0]: group for group in combination}) for combination in combinations]

        # computing crosstabs
        xtab = nan_crosstab(X[feature], y)
        xtab_test = nan_crosstab(X_test[feature], y_test)

        # printing the group statistics
        if self.verbose:
            print(f'\n - Initial distribution of {feature}')
            self.display_target_rate(feature, xtab, xtab_test)

        # measuring association with target for each combination and testing for stability on TEST
        best_groups = best_combination(combinations, self.sort_by, xtab[notna(xtab.index)], 
                                       xtab_test[notna(xtab_test.index)])

        # update of the values_orders grouped modalities in values_orders
        if best_groups:

            self.update_values_orders(feature, best_groups['combination'])
            # update of the crosstabs
            xtab = xtab.groupby(list(map(best_groups['combination'].get_group, xtab.index)), dropna=False).sum()
            xtab_test = xtab_test.groupby(list(map(best_groups['combination'].get_group, xtab_test.index)), dropna=False).sum()

        # testing adding NaNs to built groups
        order = self.values_orders.get(feature)
        if any(isna(X[feature])) & (not self.keep_nans) & (len(order) <= self.max_n_mod):

            # getting all possible combinations for the feature with NaNS
            combinations = get_all_nan_combinations(order)
            combinations = [GroupedList({group[0]: group for group in combination}) for combination in combinations]

            # adding nans at the end of the order
            order.append(nan)

            # measuring association with target for each combination and testing for stability on TEST
            best_groups = best_combination(combinations, self.sort_by, xtab, xtab_test)

            # update of the values_orders grouped modalities in values_orders
            if best_groups:

                self.update_values_orders(feature, best_groups['combination'])
                # update of the crosstabs
                xtab = xtab.groupby(list(map(best_groups['combination'].get_group, xtab.index)), dropna=False).sum()
                xtab_test = xtab_test.groupby(list(map(best_groups['combination'].get_group, xtab_test.index)), dropna=False).sum()

        # printing the new group statistics
        if self.verbose and best_groups:

            print(f'\n - Fitted distribution of {feature}')
            self.display_target_rate(feature, xtab, xtab_test, plot=True)

        return best_groups

    def update_values_orders(self, feature: str, best_groups: GroupedList):
        """ Updates the values_orders according to the best_groups"""

        # accessing current order for specified feature
        order = self.values_orders.get(feature)

        # grouping for each group of the combination
        for kept, discarded in best_groups.contained.items():
            order.group_list(discarded, kept)

    def display_target_rate(self, feature: str, xtab: DataFrame, xtab_test: DataFrame, plot: bool=False):
        """ Pretty display of frequency and target rate per modality on the same line. """

        known_order = self.values_orders.get(feature)
        stat_cols = ['target_rate', 'frequency']

        # target rate and frequency on TRAIN
        train_stats = DataFrame({
            'target_rate': xtab[1].divide(xtab[0].add(xtab[1])),  # target rate of each modality
            'frequency': xtab[1].add(xtab[0]) / xtab.sum().sum()  # frequency of each modality
        })
    
        # target rate and frequency on TEST
        test_stats = DataFrame({
            'target_rate': xtab_test[1].divide(xtab_test[0].add(xtab_test[1])),  # target rate of each modality
            'frequency': xtab_test[1].add(xtab_test[0]) / xtab_test.sum().sum()  # frequency of each modality
        })

        # default ordering based on observed target rate (on TRAIN)
        train_order = list(train_stats.sort_values('target_rate').index)
        test_order = list(test_stats.sort_values('target_rate').index)

        # keeping numpy.nan at the end of the order (default ordering)
        if any(isna(xtab.index)):
            train_order = [c for c in train_order if notna(c)]
            train_order += [nan]
            test_order = [c for c in test_order if notna(c)]
            test_order += [nan]

        # ordering by target rate
        train_stats = train_stats.reindex(train_order)
        test_stats = test_stats.reindex(test_order)

        # ordering statistics if an order already exists
        if known_order is not None:

            # known modalities
            new_train_order = known_order[:]

            # adding unexpected modalities (only present in train_order)
            new_train_order += [elt for elt in train_order if not any(is_equal(elt, t) for t in new_train_order)]

            # ordering TRAIN
            train_stats = train_stats.reindex(new_train_order)

            # adding start and end modalities
            renamed_train_order = [f'{known_order.get(c)[-1]} to {known_order.get(c)[0]}' if c in known_order and len(known_order.get(c)) > 1 else c for c in new_train_order]
            train_stats.index = renamed_train_order

            # ordering the TEST statistics based on known order
            new_test_order = new_train_order[:]

            # adding unexpected modalities (only present in test_order)
            new_test_order += [elt for elt in test_order if not any(is_equal(elt, t) for t in new_test_order)]

            # ordering TEST
            test_stats = test_stats.reindex(new_test_order)

            # adding start and end modalities
            renamed_test_order = [f'{known_order.get(c)[-1]} to {known_order.get(c)[0]}' if c in known_order and len(known_order.get(c)) > 1 else c for c in new_test_order]
            test_stats.index = renamed_test_order

        # plotting the train stats
        if plot:
        	fig, ax = plot_stats(train_stats)

        # displaying the train and test stats
        else:

            # Displaying feature level stats
            styler = concat([train_stats, test_stats], ignore_index=True).style.background_gradient(cmap='coolwarm')  # unifying colors for both stats
            train_style = train_stats.style.use(styler.export()).set_table_attributes("style='display:inline'").set_caption('Train:')
            test_style = test_stats.style.use(styler.export()).set_table_attributes("style='display:inline'").set_caption('Test:')
            display_html(train_style._repr_html_() + ' ' + test_style._repr_html_(), raw=True)


def best_combination(combinations: list, sort_by: str, xtab: DataFrame, xtab_test: DataFrame):
    """ Finds the best combination of groups of feature's values:
     - Most associated combination on train sample 
     - Stable target rate of combination on test sample.
    """

    # initiating list of associations between cut feature and target 
    associations = []

    # iterating over each combination
    for combination in combinations:
        
        # converting combination to a grouped list
        association = {'combination': combination}

        # grouping modalities in the initial crosstab
        groups = list(map(combination.get_group, xtab.index))
        combi_xtab = xtab.groupby(groups, dropna=False).sum()
        association.update({'combi_xtab': combi_xtab})

        # measuring association with the target
        association.update(association_xtab(combi_xtab))
        
        # storing results
        associations += [association]

    # sort according to association measure
    if any(combinations):
        associations = DataFrame(associations)
        associations.sort_values(sort_by, inplace=True, ascending=False)
        associations = associations.to_dict(orient='records')

    # testing associations
    for association in associations:
        
        # needed parameters
        combination, combi_xtab = association['combination'], association['combi_xtab']

        # grouping modalities in the initial crosstab
        combi_xtab_test = xtab_test.groupby(list(map(combination.get_group, xtab_test.index)), dropna=False).sum()

        # checking that all non-nan groups are in TRAIN and TEST
        unq_x =  [v for v in unique(combi_xtab.index) if v != notna(v)]
        unq_xtest = [v for v in unique(combi_xtab_test.index) if v != notna(v)]
        viability = all([e in unq_x for e in unq_xtest])
        viability = viability and all([e in unq_xtest for e in unq_x])

        # same target rate order in TRAIN and TEST
        train_target_rate = combi_xtab[1].divide(combi_xtab[0]).sort_values()
        test_target_rate = combi_xtab_test[1].divide(combi_xtab_test[0]).sort_values()
        viability = viability and all(train_target_rate.index == test_target_rate.index)

        # checking that some combinations were provided
        if viability:

        	association.update({'combi_xtab_test': combi_xtab_test})

        	return association


def get_all_combinations(values: list, q: int, max_n_mod: int=None):

    # max number of classes
    if max_n_mod is None:
    	max_n_mod = q

    # all possible combinations
    combinations = list()
    for n_class in range(2, max_n_mod + 1):
        combinations += get_combinations(n_class, q)

    combinations = [[values[int(c[0]): int(c[1]) + 1] for c in combi] for combi in combinations]  # getting real feature values

    return combinations

def get_all_nan_combinations(order: list):
    """ all possible combinations of modalities with numpy.nan"""

    # computing all non-NaN combinations
    if len(order) > 1:
        combinations = get_all_combinations(order, len(order))
    else:
        combinations = [[order]]

    # iterating over each combinations of non-NaNs
    new_combinations = []
    for combi in combinations:

         # NaNs not attributed to a group (own modality)
        new_combinations += [combi + [[nan]]] 

        # NaNs attributed to a group of non NaNs
        for n in range(len(combi)):

            # grouping NaNs with an existing group
            new_combination = [combi[n] + [nan]]

            # adding other groups unchanged
            pre = [o for o in combi[:n]]
            nex = [o for o in combi[n+1:]]
            new_combinations += [pre + new_combination + nex]

    return new_combinations


def consecutive_combinations(n_remaining: int, start: int, end: int, grp_for_this_step: list=None):
    """HELPER finds all consecutive combinations between start and end.    """

    # Import de la liste globale
    global __li_of_res__

    # initiating group
    if not grp_for_this_step:
    	grp_for_this_step = []
    
    # stopping when there are non more remaining classes
    if n_remaining == 0:
        
        ### On ajoute le dernier quantile restant au découpage ###
        grp_for_this_step += [(start, end)]

        ### Ajout du découpage réalisé au groupe des solutions ###
        __li_of_res__ += [grp_for_this_step]

    
    # adding all possible combinations of each possible range
    else:
        
        ### Parcours de toutes les valeurs possibles de fin pour le i-ème groupe du groupement à x quantiles ###
        for i in range(start, end):

            consecutive_combinations(n_remaining-1, i+1, end, grp_for_this_step + [(start, i)])
            
def get_combinations(n_class, q):
    """HELPER recupération des combinaisons possibles de q quantiles pour n_class."""
    
    globals()['__li_of_res__'] = []

    consecutive_combinations(n_class-1, 0, q-1)
    
    combinations = [list(map(lambda u: (str(u[0]).zfill(len(str(q-1))), str(u[1]).zfill(len(str(q-1)))), l)) for l in __li_of_res__]
    
    return combinations

def association_xtab(xtab: DataFrame):
    """ Computes measures of association between feature x and feature2. """

    # numnber of observations
    n_obs = xtab.sum().sum()

    # number of values taken by the features
    n_mod_x, n_mod_y = len(xtab.index), len(xtab.columns)

    # Chi2 statistic
    chi2 = chi2_contingency(xtab)[0]

    # Cramer's V
    cramerv = sqrt(chi2 / n_obs)

    # Tschuprow's T
    n_mods = sqrt((n_mod_x - 1) * (n_mod_y - 1))
    tschuprowt = 0
    if n_mods > 0:
        tschuprowt = sqrt(chi2 / n_obs / n_mods)

    results = {'cramerv': cramerv, 'tschuprowt': tschuprowt}
    
    return results

def nan_crosstab(x: Series, y: Series):
    """ Crosstab that keeps nans as a specific value"""
    
    # keeping NaNs as a specific modality
    nan_val = '__NAN__'  # value to be imputed as NaN
    x_filled = x.fillna(nan_val)  # filling NaNs

    # computing initial crosstabs
    xtab = crosstab(x_filled, y)

    # keeping track of nans if there are any
    xtab.index = [idx if idx != nan_val else nan for idx in xtab.index]
    
    return xtab


def plot_stats(stats):
    """ Barplot of the volume and target rate"""
    
    x = [0] + [elt for e in stats['frequency'].cumsum()[:-1] for elt in [e] * 2] + [1]
    y2 = [elt for e in list(stats['target_rate']) for elt in [e]*2]
    s = list(stats.index)
    c = color_palette("coolwarm", as_cmap=True)(y2)

    fig, ax = subplots()

    for i in range(len(stats)):
        k = i*2
        ax.fill_between(x[k: k+2], [0, 0], y2[k: k+2], color=c[k])
        ax.text(sum(x[k: k+2]) / 2, y2[k], s[i], ha='center', va='bottom')

    ax.xaxis.set_major_formatter(PercentFormatter(xmax=1))    
    ax.yaxis.set_major_formatter(PercentFormatter(xmax=1))    
    ax.set_xlabel('Volume')
    ax.set_ylabel('Target rate')
    ax.set_xlim(0, 1)
    ax.set_ylim(0, 1)
    despine()
    
    return fig, ax
